//N.Yu.Zolotykh 1999
//University of Nizhni Novgorod, Russia
#include <stdlib.h>
#include "matrices.h"
#include "powerest.h"

#ifndef GCD_H_
#define GCD_H_

/**
    \file
   Euclidian algorithm.
   This module consists of two functions which implements Euclidian algorithm
   for greatest common divisor producing.
*/


/**
   Euclidian algorithm produces
   the greatest common divisor of two integer u and v
 */

template <class item_class>
  item_class gcd(const item_class & a, const item_class & b)
{
  item_class r;
  item_class u = absolute_value(a);
  item_class v = absolute_value(b);
  while (v != 0)
  {
    r = u % v;
    u = v;
    v = r;
  }
  return u;
}


/**
   Euclidian algorithm produces
   the greatest common divisor of two integer u and v
 */

template <class T>
  T gcd_bezout(const T & a, const T & b, T & u, T & v)
{
    u     = 1;
  T u_new = 0;

    v     = 0;
  T v_new = 1;

  T r 	  = a;
  T r_new = b;

  while (r_new != (T)0)
  {
    T q = quotient (r, r_new);

    T aux = u_new;
    u_new = u - q * u_new;
    u     = aux;

    aux   = v_new;
    v_new = v - q * v_new;
    v     = aux;

    aux   = r_new;
    r_new = r - q * r_new;
    r     = aux;
  }
  return r;
}


  /**
     Euclidian algorithm that produces
     the greatest common divisor of entries in v
   */
template <class item_class>
  item_class gcd(const vector<item_class> & v)
{
  item_class d = v.at(v.get_n() - 1);
  index j = v.get_n() - 2;
  while ((d != 1) && (j >= 0))
  {
    d = gcd(v.at(j), d);
    j--;
  }
  return d;
}

#endif
